from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from bclib.parser.answer.answer import Answer
from ..answer.user_action_types import UserActionTypes
from ..answer.file_user_action import FileUserAction


class UserAction:
    """This class will be updated in next versions."""

    def __init__(self, prp_id: 'int', action: 'UserActionTypes', prp_value_id: 'int', internal_prp_value_id: 'int', value_id: 'int', value: 'any', datatype: 'str',
                 multi: 'bool', part: 'int', answer: 'Answer'):
        self.prp_id = prp_id
        self.action = action
        self.prp_value_id = prp_value_id
        self.value_id = value_id
        self.value = value
        self.part = part
        self.datatype = datatype
        self.multi = multi
        self.answer: 'Answer' = answer
        self.internal_prp_value_id: "int" = internal_prp_value_id
        self.database:"str" = None
        self.table:"str" = None
        self.field:"str" = None
        self.ownerid:"int" = 0
        self.typeid:"int" = None
        self.wordid:"int" = None
        self.validation_status: "bool" = True
        self.validation_message: "list[str]" = []

    def as_tuple(self) -> tuple:
        return (
            self.prp_id, self.action.value, self.prp_value_id, self.internal_prp_value_id, self.value_id,
            self.value, self.part, self.datatype, self.database, self.table, self.field, self.multi, 
            self.ownerid, self.typeid, self.wordid, self.answer, self.validation_status, self.validation_message
        )

    def as_dict(self) -> dict:
        return {
            "prp_id": self.prp_id,
            "action": self.action.value,
            "prp_value_id": self.prp_value_id,
            "internal_prp_value_index": self.internal_prp_value_id,
            "value_id": self.value_id,
            "value": self.value,
            "part": self.part,
            "datatype": self.datatype,
            "database": self.database,
            "table": self.table,
            "field": self.field,
            "multi": self.multi,
            "ownerid": self.ownerid,
            "typeid": self.typeid,
            "wordid": self.wordid,
            "answer": self.answer,
            "validation_status": self.validation_status,
            "validation_message": self.validation_message
        }

    def is_file_content(self):
        return self.datatype == "files"

    def as_file_content(self) -> 'FileUserAction':
        if self.is_file_content():
            value = self.value
            return FileUserAction(self.prp_id, self.prp_value_id, self.value_id, value["name"], value["type"], value["size"], value["content"], value.get("uploadToken"))

    def __str__(self) -> str:
        return str(self.as_tuple())

    def __repr__(self) -> str:
        return str(self.as_dict())



import asyncio
import json
from typing import Any
from bclib.listener.message_type import MessageType
from abc import abstractmethod


class Message:
    def __init__(self, sessionId: str,  messageType: MessageType, buffer: bytes = None) -> None:
        self.session_id = sessionId
        self.type = messageType
        self.buffer = buffer

    @abstractmethod
    def create_response_message(self, session_id: str, buffer: bytes) -> "Message":
        return Message.create_add_hock(session_id,buffer)


    async def write_to_stream_async(self, stream: asyncio.StreamWriter) -> bool:
        is_send = True
        try:
            stream.write(self.type.value.to_bytes(1, 'big'))
            data = self.session_id.encode()
            data_length_bytes = len(data).to_bytes(4, 'big')
            stream.write(data_length_bytes)
            stream.write(data)

            if self.type in (MessageType.AD_HOC, MessageType.MESSAGE):
                data_length_bytes = len(self.buffer).to_bytes(4, 'big')
                stream.write(data_length_bytes)
                stream.write(self.buffer)
            await stream.drain()
        except asyncio.CancelledError:
            is_send = False
        return is_send

    @staticmethod
    def create_add_hock(session_id: str, buffer: bytes):
        return Message(session_id, MessageType.AD_HOC, buffer)

    @staticmethod
    def create_disconnect(session_id: str):
        return Message(session_id, MessageType.DISCONNECT, None)

    @staticmethod
    def create_from_text(session_id: str, text: str):
        return Message(session_id, MessageType.MESSAGE, text.encode())

    @staticmethod
    def create_from_byte(session_id: str, array: bytes):
        return Message(session_id, MessageType.MESSAGE, array)

    @staticmethod
    def create_from_object(session_id: str, object_data: Any):
        return Message(session_id, MessageType.MESSAGE, json.dumps(object_data).encode("utf-8"))

    @staticmethod
    def create(session_id: str, data: Any):
        ret_val: Message = None
        if isinstance(data, str):
            ret_val = Message.create_from_text(
                session_id,  data)
        elif isinstance(data, bytes):
            ret_val = Message.create_from_byte(
                session_id,  data)
        else:
            ret_val = Message.create_from_object(
                session_id,  data)
        return ret_val



from types import FunctionType
from bclib.context import Context
from bclib.utility import DictEx
from ..predicate.predicate import Predicate


class Url (Predicate):
    """Create Url cheking predicate"""

    def __init__(self, expression: str) -> None:
        super().__init__(expression)
        self.__validator: FunctionType = Url.__generate_validator(expression)

    async def check_async(self, context: Context) -> bool:
        try:
            is_ok, url_parts = self.__validator(context.url)
            if is_ok and url_parts:
                context.url_segments = DictEx(url_parts)
            return is_ok
        except Exception as ex:
            print("Error in check url predicate", ex)
            return False

    @staticmethod
    def __generate_validator(url: str) -> FunctionType:
        segment_list = []
        return_dict_property_names = []
        where_part_list = []
        parts = url.split("/")
        last_part_index = len(parts)-1
        for index, value in enumerate(parts):
            name = "_"
            if len(value) > 0 and value[0] == ':':
                name = value[1:]
                if index == last_part_index and name[0] == '*':
                    name = name[1:]
                    return_dict_property_names.append(
                        f'"{name}" : "/".join(__{name})')
                    name = f"*__{name}"
                else:
                    return_dict_property_names.append(
                        f'"{name}" : __{name}')
                    name = f"__{name}"
            else:
                where_part_list.append(
                    f"url_parts[{index}].lower() == '{value.lower()}'")
            segment_list.append(name)
        if len(where_part_list) == 0:
            if len(return_dict_property_names) == 0:
                where_part_list.append("True")
            else:
                where_part_list.append(
                    f"len(url_parts) == {len(return_dict_property_names)}")
                if len(return_dict_property_names) == 1:
                    where_part_list.append("len(url_parts[0]) > 0")

        if len(return_dict_property_names) > 0:
            body = f"""
def url_function(url):
    try:
        url_parts = url.split("/")
        if {" and ".join(where_part_list)}:
            {','.join(segment_list)} = url_parts{"[0]" if len(segment_list)==1 else ""}
            return (True,{{ {','.join(return_dict_property_names)} }})
        else:
            return (False,None)
    except Exception as e:
        return (False,None)"""
        else:
            body = f"""
def url_function(url):
    return (url.lower() == '{url.lower()}' ,None)"""
        f_code = compile(body, "<str>", "exec")
        f_func = FunctionType(f_code.co_consts[0], globals(), "url_function")
        return f_func



import asyncio
import cgi
import io
import datetime
import json
import uuid
import ssl
import base64
from typing import Callable, TYPE_CHECKING, Optional, Awaitable
from urllib.parse import unquote, parse_qs

from bclib.listener.message_type import MessageType
from ..endpoint import Endpoint
from ..http_listener.http_base_data_name import HttpBaseDataName
from ..http_listener.http_base_data_type import HttpBaseDataType
from bclib.utility import DictEx, ResponseTypes
from ..message import Message
from ..web_message import WebMessage
import pathlib

if TYPE_CHECKING:
    from aiohttp import web

from aiohttp.log import web_logger

class HttpListener:
    _id = 0
    LOGGER = "logger"
    ROUTER = "router"
    MIDDLEWARES = "middlewares"
    HANDLER_ARGS = "handler_args"
    CLIENT_MAX_SIZE = "client_max_size"

    _DEFAULT_LOGGER = web_logger
    _DEFAULT_ROUTER = None
    _DEFAULT_MIDDLEWARES = ()
    _DEFAULT_HANDLER_ARGS = None
    _DEFAULT_CLIENT_MAX_SIZE = 1024 ** 2
    

    def __init__(self, endpoint: Endpoint, async_callback: 'Callable[[WebMessage], Awaitable[WebMessage]]',ssl_options:'dict', configuration: Optional[DictEx]):
        self.__endpoint = endpoint
        self.on_message_receive_async = async_callback
        self.ssl_options = ssl_options
        self.__config = configuration if configuration is not None else DictEx()
        self.__logger = self.__config.get(HttpListener.LOGGER, HttpListener._DEFAULT_LOGGER)
        self.__router = self.__config.get(HttpListener.ROUTER, HttpListener._DEFAULT_ROUTER)
        self.__middlewares = self.__config.get(HttpListener.MIDDLEWARES, HttpListener._DEFAULT_MIDDLEWARES)
        self.__handler_args = self.__config.get(HttpListener.HANDLER_ARGS, HttpListener._DEFAULT_HANDLER_ARGS)
        self.__client_max_size = self.__config.get(HttpListener.CLIENT_MAX_SIZE, HttpListener._DEFAULT_CLIENT_MAX_SIZE)
        

    def initialize_task(self, event_loop: asyncio.AbstractEventLoop):
        event_loop.create_task(self.__server_task(event_loop))

    async def __server_task(self, event_loop: asyncio.AbstractEventLoop):
        from aiohttp import web
        from multidict import MultiDict

        async def uptime_handler(request):
            # http://HOST:PORT/?interval=90
            interval = 1;#int(request.GET.get('interval', 1))
            
            # Without the Content-Type, most (all?) browsers will not render 
            # partially downloaded content. Note, the response type is 
            # StreamResponse not Response.
            resp = web.StreamResponse(status=200, 
                                    reason='OK', 
                                    headers={'Content-Type': 'text/html; charset=utf-8'})
            
            # The StreamResponse is a FSM. Enter it with a call to prepare.
            await resp.prepare(request)
            
        
            i=3
            while i>0:
                i=i-1
                try:
                    print("hi " + str(i))
                    # Technically, subprocess blocks, so this is a dumb call 
                    # to put in an async example. But, it's a tiny block and 
                    # still mocks instantaneous for this example.
                    await resp.write(b"<strong>")
                    await resp.write(str(i).encode())#subprocess.check_output('uptime'))
                    await resp.write(b"</strong>sdsdsds<br/>\n")
                    
                    # Yield to the scheduler so other processes do stuff.
                    await resp.drain()
                    
                    # This also yields to the scheduler, but your server 
                    # probably won't do something like this. 
                    await asyncio.sleep(interval)
                except Exception as e:
                    # So you can observe on disconnects and such.
                    print(repr(e))
                    raise
            
            return resp
        
        async def on_request_receive_async(request: 'web.Request') -> web.Response:
            ret_val: web.Response = None
            request_cms = await self.create_cms_async(request)
            msg = WebMessage(request, str(uuid.uuid4()),MessageType.AD_HOC,json.dumps(request_cms, ensure_ascii=False).encode(encoding="utf-8"))
            result = await self.on_message_receive_async(msg)
            if result and result.Response is None:
                cms: dict = json.loads(result.buffer.decode("utf-8"))
                cms_cms = cms[HttpBaseDataType.CMS]
                cms_cms_webserver = cms_cms[HttpBaseDataType.WEB_SERVER]
                index = cms_cms_webserver[HttpBaseDataName.INDEX]
                header_code: str = cms_cms_webserver[HttpBaseDataName.HEADER_CODE]
                mime = cms_cms[HttpBaseDataName.WEB_SERVER][HttpBaseDataName.MIME]
                headers = MultiDict()
                if HttpBaseDataName.HTTP in cms_cms:
                    http: dict = cms_cms[HttpBaseDataName.HTTP]
                    for key, value in http.items():
                        if isinstance(value, list):
                            for item in value:
                                headers.add(key, item)
                        else:
                            headers.add(key, value)
                headers.add("Content-Type", mime)
                if index == ResponseTypes.STATIC_FILE:
                    try:
                        path = pathlib.Path(cms_cms_webserver[HttpBaseDataName.FILE_PATH])
                        path.stat()
                        ret_val = web.FileResponse(
                            path=path,
                            chunk_size=256*1024,
                            status=int(header_code.split(' ')[0]),
                            headers=headers
                        )
                    except FileNotFoundError:
                        ret_val = web.Response(
                            status=404,
                            reason="File not found"
                        )
                else:
                    ret_val = web.Response(
                        status=int(header_code.split(' ')[0]),
                        headers=headers
                    )
                    if HttpBaseDataName.CONTENT in cms_cms:
                        ret_val.text = cms_cms[HttpBaseDataName.CONTENT]
                    else:
                        raw_blob_content = cms_cms[HttpBaseDataName.BLOB_CONTENT]
                        ret_val.body = base64.b64decode(raw_blob_content.encode("utf-8"))
            else:
                ret_val =  web.Response() if result.Response is None else result.Response
            return ret_val

        app = web.Application(
            logger=self.__logger,
            router=self.__router,
            middlewares=self.__middlewares,
            handler_args=self.__handler_args,
            client_max_size=self.__client_max_size,
            loop=event_loop
        )
        app.add_routes(
            [web.route('*', '/{tail:.*}', on_request_receive_async)])
        ssl_context= None
        if self.ssl_options:
            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            if 'certfile' in self.ssl_options:
                ssl_context.load_cert_chain(
                    certfile=self.ssl_options.certfile,
                    keyfile=self.ssl_options.keyfile
                )
            elif 'pfxfile' in self.ssl_options:
                try:
                    pem_file_path = HttpListener.convert_pfx_to_pem_file(
                        self.ssl_options.pfxfile,
                        self.ssl_options.password
                    )
                    ssl_context.load_cert_chain(certfile=pem_file_path)
                except Exception as e:
                    raise Exception("Invalid PKCS12 or pastphrase for {0}: {1}".format(self.ssl_options.pfxfile, e))

        runner = web.AppRunner(app, handle_signals=True)
        await runner.setup()
        site = web.TCPSite(runner, self.__endpoint.url, self.__endpoint.port, ssl_context=ssl_context)
        await site.start()
        print(
            f"Development Edge server started at http{'s' if self.ssl_options else ''}://{self.__endpoint.url}:{self.__endpoint.port}")
        try:
            while True:
                await asyncio.Future()
        except asyncio.CancelledError:
            pass
        finally:
            print(f"Development Edge server for http{'s' if self.ssl_options else ''}://{self.__endpoint.url}:{self.__endpoint.port} stopped.")
            await site.stop()
            await runner.cleanup()
            await runner.shutdown()

    @staticmethod
    def convert_pfx_to_pem_file(pfxfile:str,password:str)->str:
        from cryptography.hazmat.primitives.serialization import pkcs12, Encoding, PrivateFormat, NoEncryption
        with open(pfxfile,"rb") as f:
            try:
                private_key, certificate, additional_certificates = pkcs12.load_key_and_certificates(f.read(), password.encode())
                pem_file_path = '{0}.auto-generated.pem'.format(pfxfile)
                with open(pem_file_path, 'wb') as pem_file:
                    pem_file.write(certificate.public_bytes(Encoding.PEM))
                    for item in additional_certificates:
                        pem_file.write(item.public_bytes(Encoding.PEM))
                    pem_file.write(private_key.private_bytes(encoding= Encoding.PEM,format=PrivateFormat.TraditionalOpenSSL,encryption_algorithm=NoEncryption()))
                return pem_file_path
            except Exception as ex:
                raise Exception("Error in create pem file from pfx {0}: {1}".format(pfxfile, ex))

    @staticmethod
    async def create_cms_async(request: 'web.Request') -> dict:
        cms_object = dict()
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.METHODE, request.method.lower())
        raw_url = unquote(request.path_qs)[1:]
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.RAW_URL, raw_url)
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.URL, request.path[1:])
        HttpListener.__add_query_string(request.query, cms_object)
        for key, value in request.headers.items():
            field_name = key.strip().lower()
            if field_name == HttpBaseDataName.COOKIE:
                HttpListener.__add_cookie(value, cms_object)
            elif field_name == HttpBaseDataName.HOST:
                HttpListener.__add_host(value, raw_url, cms_object)
            else:
                HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                          field_name, str(value).strip())
        HttpListener.__add_server_data(cms_object, request)
        await HttpListener.__add_body_async(cms_object, request)
        return {"cms": cms_object}

    @staticmethod
    async def __add_body_async(cms_object: dict, request: 'web.Request'):
        content_len_str = request.headers.get('Content-Length')
        if content_len_str or request.can_read_body:
            raw_body = await request.read()
            body = raw_body.decode('utf-8')
            HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                      HttpBaseDataName.BODY, body)
            content_type: str = request.headers.get(
                "content-type")
            if content_type and content_type.find("application/json") < 0:
                if content_type.find("multipart/form-data") >= 0:
                    _, content_type_value_params = cgi.parse_header(
                        content_type)
                    content_type_value_params['boundary'] = bytes(
                        content_type_value_params['boundary'], "utf-8")
                    if content_len_str is not None:
                        content_type_value_params['CONTENT-LENGTH'] = int(
                            content_len_str)
                    with io.BytesIO(raw_body) as stream:
                        fields = cgi.parse_multipart(
                            stream, content_type_value_params)
                    for key, value in fields.items():
                        HttpListener.__add_header(cms_object,
                                                  HttpBaseDataType.FORM, key, value[0] if len(value) == 1 else value)
                else:
                    for key, value in parse_qs(body).items():
                        HttpListener.__add_header(
                            cms_object, HttpBaseDataType.FORM, key, value[0] if len(value) == 1 else value)

    @staticmethod
    def __add_server_data(cms_object: dict, request: 'web.Request'):
        HttpListener._id += 1
        now = datetime.datetime.now()
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.DATE, now.strftime("%d/%m/%Y"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.TIME, now.strftime("%H:%M"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.DATE2, now.strftime("%Y%m%d"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.TIME2, now.strftime("%H%M%S"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.DATE3, now.strftime("%Y.%m.%d"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.REQUEST_ID, str(HttpListener._id))
        host_parts = request.host.split(':')
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.HOST_IP, host_parts[0])
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.HOST_PORT,  host_parts[1] if len(host_parts) > 1 else "80")  # edit
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.CLIENT_IP, str(request.remote))

    @staticmethod
    def __add_query_string(query: dict, cms_object: dict) -> None:
        for key, value in query.items():
            HttpListener.__add_header(
                cms_object, HttpBaseDataType.QUERY, key, value)

    @staticmethod
    def __add_cookie(raw_header_value: str, cms_object) -> None:
        for item in raw_header_value.split(';'):
            parts = item.split('=')
            if len(parts) == 2:
                HttpListener.__add_header(cms_object, HttpBaseDataName.COOKIE,
                                          parts[0].strip(), parts[1].strip())

    @staticmethod
    def __add_host(row_host: str, row_url: str, cms_object) -> None:
        host_parts = row_host.split(':')
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.HOST, host_parts[0])
        if len(host_parts) == 2:
            HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                      HttpBaseDataName.PORT, host_parts[1])
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.FULL_URL,  f"{row_host}/{row_url}")

    @staticmethod
    def __add_header(cms_object: dict, value_type: str, value_name: str, value: str) -> None:
        if value_type not in cms_object:
            cms_object[value_type] = dict()
        type_node = cms_object[value_type]
        if value_name in type_node:
            name_node = type_node[value_name]
            if isinstance(name_node, list):
                name_node.append(value)
            else:
                type_node[value_name] = [name_node, value]
        else:
            type_node[value_name] = value



from abc import ABC
import traceback
from typing import TYPE_CHECKING, Tuple

from bclib.exception import ShortCircuitErr

from bclib.db_manager import SqlDb, SQLiteDb, MongoDb, RabbitConnection, RESTfulConnection
from bclib.utility import DictEx, HttpStatusCodes, HttpStatusCodes
from bclib.listener.http_listener import HttpBaseDataName, HttpBaseDataType
import base64

if TYPE_CHECKING:
    from .. import dispatcher


class Context(ABC):
    """Base class for dispatching"""

    def __init__(self, dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__()
        self.dispatcher = dispatcher
        self.url_segments: DictEx = None
        self.url: str = None
        self.is_adhoc = True

    def open_sql_connection(self, key: str) -> SqlDb:
        return self.dispatcher.db_manager.open_sql_connection(key)

    def open_sqllite_connection(self, key: str) -> SQLiteDb:
        return self.dispatcher.db_manager.open_sqllite_connection(key)

    def open_mongo_connection(self, key: str) -> MongoDb:
        return self.dispatcher.db_manager.open_mongo_connection(key)

    def open_restful_connection(self, key: str) -> RESTfulConnection:
        return self.dispatcher.db_manager.open_restful_connection(key)

    def open_rabbit_connection(self, key: str) -> RabbitConnection:
        return self.dispatcher.db_manager.open_rabbit_connection(key)

    def generate_error_response(self, exception: Exception) -> dict:
        """Generate error response from process result"""

        error_object, _ = self._generate_error_object(exception)
        return error_object

    def _generate_error_object(self, exception: Exception) -> 'Tuple[dict, str]':
        """Generate error object from exception object"""
        error_code = None
        data = None
        status_code = HttpStatusCodes.INTERNAL_SERVER_ERROR
        if isinstance(exception, ShortCircuitErr):
            data = exception.data
            status_code = exception.status_code
            error_code = exception.error_code
        if data:
            error_object = data
        else:
            error_object = {
                "errorCode": error_code,
                "errorMessage": str(exception)
            }
            if self.dispatcher.log_error:
                error_object["error"] = traceback.format_exc()
        return (error_object, status_code)

    @staticmethod
    def _generate_response_cms(
            content: 'str|bytes',
            response_type: 'str',
            status_code: 'str',
            mime: 'str',
            template: 'DictEx' = None,
            headers: 'dict' = None) -> dict:
        """Generate response from process result"""

        ret_val = DictEx() if template is None else template
        if HttpBaseDataType.CMS not in ret_val:
            ret_val[HttpBaseDataType.CMS] = {}
        if HttpBaseDataName.WEB_SERVER not in ret_val[HttpBaseDataType.CMS]:
            ret_val[HttpBaseDataType.CMS][HttpBaseDataName.WEB_SERVER] = DictEx()
        ret_val[HttpBaseDataType.CMS][HttpBaseDataName.WEB_SERVER][HttpBaseDataName.INDEX] = response_type
        ret_val[HttpBaseDataType.CMS][HttpBaseDataName.WEB_SERVER][HttpBaseDataName.HEADER_CODE] = status_code
        ret_val[HttpBaseDataType.CMS][HttpBaseDataName.WEB_SERVER][HttpBaseDataName.MIME] = mime
        if isinstance(content,bytes):
            ret_val[HttpBaseDataType.CMS][HttpBaseDataName.BLOB_CONTENT] = base64.b64encode(content).decode("utf-8")  
        else:
            ret_val[HttpBaseDataType.CMS][HttpBaseDataName.CONTENT] = content
        if headers is not None:
            Context.__add_user_defined_headers(ret_val, headers)
        return ret_val

    @staticmethod
    def __add_user_defined_headers(response: dict, headers: dict) -> None:
        """Adding user defined header to response"""

        if HttpBaseDataName.HTTP not in response[HttpBaseDataType.CMS]:
            response[HttpBaseDataType.CMS][HttpBaseDataName.HTTP] = {}
        http = response[HttpBaseDataType.CMS][HttpBaseDataName.HTTP]
        for key, value in headers.items():
            if key in http:
                current_value = http[key] if isinstance(
                    http[key], list) else [http[key]]
                new_value = current_value + value
            else:
                new_value = value

            http[key] = ",".join(new_value)



import asyncio
import json
import uuid
from typing import Any
from ..named_pipe.inamed_pipe_connection import INamedPipeConnection

from bclib.utility import WindowsNamedPipeHelper


class WindowsNamedPipeConnection(INamedPipeConnection):
    def __init__(self, pipe_name: str, event_loop: asyncio.AbstractEventLoop) -> None:
        self.__event_loop = event_loop
        self.__reader_pipe_name = F"{pipe_name}/reader"
        self.__writer_pipe_name = F"{pipe_name}/writer"
        self.__reader_pipe_handle = None
        self.__writer_pipe_handle = None
        self.__query_list: 'dict[str,asyncio.Future]' = dict()
        self.__propcess_tesk = self.__event_loop.create_task(
            self.__get_receive_message_Task())

    def __connect_to_reader_named_pipe(self):
        import win32file
        import pywintypes
        try:
            self.__reader_pipe_handle = win32file.CreateFile(self.__reader_pipe_name, win32file.GENERIC_WRITE,
                                                             0, None, win32file.OPEN_EXISTING, win32file.FILE_ATTRIBUTE_NORMAL, None)
        except pywintypes.error as e:  # pylint: disable=maybe-no-member
            if e.args[0] == 2:   # ERROR_FILE_NOT_FOUND
                print(f"No Named Pipe '{self.__reader_pipe_name}'. {repr(e)}")
            else:
                print(
                    f"Named Pipe '{self.__reader_pipe_name}' error code {e.args[0]}. {repr(e)}")
            raise

    def __connect_to_writer_pipe(self):
        import win32file
        import pywintypes
        try:
            self.__writer_pipe_handle = win32file.CreateFile(self.__writer_pipe_name, win32file.GENERIC_READ,
                                                             0, None, win32file.OPEN_EXISTING, win32file.FILE_ATTRIBUTE_READONLY, None)

        except pywintypes.error as e:  # pylint: disable=maybe-no-member
            if e.args[0] == 2:   # ERROR_FILE_NOT_FOUND
                print(f"No Named Pipe '{self.__writer_pipe_name}'. {repr(e)}")
            else:
                print(
                    f"Named Pipe '{self.__writer_pipe_name}' error code {e.args[0]}. {repr(e)}")
            raise

    async def __get_receive_message_Task(self):
        while True:
            try:
                if self.__writer_pipe_handle is None:
                    self.__connect_to_writer_pipe()
                message = await WindowsNamedPipeHelper.read_from_named_pipe_async(self.__writer_pipe_handle, self.__event_loop)
                if message and message.session_id in self.__query_list:
                    future = self.__query_list[message.session_id]
                    del self.__query_list[message.session_id]
                    data = json.loads(message.buffer)
                    future.get_loop().call_soon_threadsafe(future.set_result, data)
            except asyncio.CancelledError:
                self.clear_query_list()
                break
            except:
                self.__writer_pipe_handle = None
                self.clear_query_list()
                await asyncio.sleep(1)

    def __try_send_command(self, command: Any) -> str:
        from bclib.listener import Message
        session_id = None
        try:
            if self.__reader_pipe_handle is None:
                self.__connect_to_reader_named_pipe()
            session_id = str(uuid.uuid4())
            msg = Message.create_add_hock(
                session_id, json.dumps(command).encode("utf-8"))
            WindowsNamedPipeHelper.write_to_named_pipe(
                msg, self.__reader_pipe_handle)
        except:
            self.__reader_pipe_handle = None
        return session_id

    def try_send_command(self, command: Any) -> bool:
        return True if self.__try_send_command(command) else False

    async def try_send_query_async(self, query: 'Any') -> 'Any':
        ret_val = None
        session_id = self.__try_send_command(query)
        if session_id:
            future = self.__event_loop.create_future()
            self.__query_list[session_id] = future
            ret_val = await future
        return ret_val

    def clear_query_list(self):
        for session_id in self.__query_list:
            try:
                future = self.__query_list[session_id]
                future.get_loop().call_soon_threadsafe(
                    future.set_exception, Exception("Named Pipe is broken"))
            except:
                pass
        self.__query_list.clear()



from bclib.db_manager.db_manager import DbManager
from bclib.db_manager.sql_db import SqlDb
from bclib.db_manager.sqlite_db import SQLiteDb
from bclib.db_manager.mongo_db import MongoDb
from bclib.db_manager.rabbit_connection import RabbitConnection
from bclib.db_manager.restful_connection import RESTfulConnection
from bclib.db_manager.named_pipe.named_pipe_connection import NamedPipeConnection
from bclib.db_manager.named_pipe.inamed_pipe_connection import INamedPipeConnection
from bclib.db_manager.named_pipe.windows_named_pipe_connection import WindowsNamedPipeConnection



class HttpBaseDataType:
    WEB_SERVER = "webserver"
    CMS = "cms"
    REQUEST = "request"
    COOKIE = "cookie"
    FORM = "form"
    QUERY = "query"
    JSON = "json"
    HEADER = "header"
    HTTP = "http"



from typing import Any, TYPE_CHECKING, Coroutine, Optional

from bclib.utility import HttpMimeTypes
from ..context.request_context import RequestContext

if TYPE_CHECKING:
    from .. import dispatcher
    from .. import listener


class WebContext(RequestContext):
    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher',message_object: 'listener.WebMessage') -> None:
        super().__init__(cms_object, dispatcher)
        self.process_async = True
        self.mime = HttpMimeTypes.HTML
        self.__message = message_object
    
    async def start_stream_response_async(self,status: int = 200,
        reason: Optional[str] = 'OK',
        headers: Optional[dict] = None,)-> Coroutine[Any, Any, None]:
        await self.__message.start_stream_response_async(status,reason,headers)

    async def write_async(self,data:'bytes') -> Coroutine[Any, Any, None]:
        await  self.__message.write_async(data)

    async def drain_async(self) -> Coroutine[Any, Any, None]:
        await self.__message.drain_async()
    
    async def write_and_drain_async(self,data:'bytes') -> Coroutine[Any, Any, None]:
        await self.write_async(data)
        await self.drain_async()



from bclib.cache.cache_item.base_cache_item import BaseCacheItem
from ..value_item.base_value_item import BaseValueItem

class ScalarValueItem(BaseValueItem):
    
    def _apply_item(self, cache_item: "BaseCacheItem") -> "BaseCacheItem":
        self._item = cache_item



"""Main module of bclib.wrapper for all exist module that need in basic coding"""

import asyncio
from bclib.db_manager import *
from bclib.dispatcher import RoutingDispatcher, IDispatcher, SocketDispatcher, DevServerDispatcher, NamedPipeDispatcher, EndpointDispatcher
from bclib.context import Context, WebContext, SocketContext, ClientSourceContext, ClientSourceMemberContext, RabbitContext, RESTfulContext, RequestContext, MergeType, ServerSourceContext, ServerSourceMemberContext, SourceContext, SourceMemberContext, NamedPipeContext
from bclib.utility import DictEx, HttpStatusCodes, HttpMimeTypes, ResponseTypes, HttpHeaders, WindowsNamedPipeHelper
from bclib.listener import Message, MessageType, HttpBaseDataType, HttpBaseDataName
from bclib.predicate import Predicate
from bclib.exception import *
from bclib import __version__


def from_config(option_file_path: str, file_name: str = "host.json"):
    """Create related RoutingDispatcher obj from config file in related path"""
    import json
    from pathlib import Path

    with open(Path(option_file_path).with_name(file_name), encoding="utf-8") as options_file:
        options = json.load(options_file)
    from_options(options)


def from_list(hosts: 'dict[str,list[str]]'):
    """Create related RoutingDispatcher obj from path list object"""
    import subprocess
    import asyncio

    __print_splash(True)
    loop = asyncio.get_event_loop()
    for host, args in hosts.items():
        try:
            args.append(f"-n {host}")
            args.append("-m")
            loop.run_in_executor(None, subprocess.run, args)
            print(f'{host} start runing from {args[1]}')
        except Exception as ex:
            print(ex)


def from_options(options: dict,loop:asyncio.AbstractEventLoop = None) -> RoutingDispatcher:
    """Create related RoutingDispatcher obj from config object"""

    import sys
    import getopt

    multi: bool = False
    argumentList = sys.argv[1:]
    # Options
    short_options = "mn:"
    # Long options
    long_options = ["Name =", "Multi"]
    try:
        arguments, _ = getopt.gnu_getopt(
            argumentList, short_options, long_options)
        for current_argument, current_value in arguments:
            if current_argument in ("-n", "--Name"):
                options["name"] = current_value.strip()
            elif current_argument in ("-m", "--Multi"):
                multi = True
    except getopt.error as err:
        print(str(err))

    if not multi:
        __print_splash(False)
    ret_val: RoutingDispatcher = None
    if "server" in options:
        ret_val = DevServerDispatcher(options=options,loop=loop)
    elif "named_pipe" in options:
        ret_val = NamedPipeDispatcher(options=options,loop=loop)
    elif "endpoint" in options:
        ret_val = EndpointDispatcher(options=options,loop=loop)
    else:
        ret_val = SocketDispatcher(options=options,loop=loop)
    return ret_val


def __print_splash(inMultiMode: bool):
    print(f'''
______           _                          _____    _            
| ___ \\         (_)                        |  ___|  | |           
| |_/ / __ _ ___ _ ___  ___ ___  _ __ ___  | |__  __| | __ _  ___ 
| ___ \\/ _` / __| / __|/ __/ _ \\| '__/ _ \\ |  __|/ _` |/ _` |/ _ \\
| |_/ / (_| \\__ \\ \\__ \\ (_| (_) | | |  __/ | |__| (_| | (_| |  __/
\\____/ \\__,_|___/_|___/\\___\\___/|_|  \\___| \\____/\\__,_|\\__, |\\___|
                                                        __/ |     
                                                       |___/      
***********************************
Basiscore Edge

Welcome To BasisCore Ecosystem
Follow us on https://BasisCore.com/
bclib Version : {__version__}
Run in {'multi' if inMultiMode else 'single'} instance mode!
***********************************
(Press CTRL+C to quit)
''')



from abc import ABC, abstractmethod
from ..cache_item.base_cache_item import BaseCacheItem

class BaseValueItem(ABC):
    def __init__(self, cache_item:"BaseCacheItem") -> None:
        super().__init__()
        self._item:"list[BaseCacheItem]|BaseCacheItem" = None
        self._apply_item(cache_item)
    
    @abstractmethod
    def _apply_item(self, cache_item:"BaseCacheItem"): ...

    def add_or_update_item(self, cache_item: "BaseCacheItem"):
        self._apply_item(cache_item)
    
    def get_item(self) -> "list|any|None":
        if self._item is not None:
            if isinstance(self._item, list):
                ret_val = list()
                for item in self._item:
                    data = item.data()
                    if data is not None:
                        ret_val.append(data)
                return ret_val if len(ret_val) > 0 else None
            else:
                ret_val = self._item.data()
        else:
            ret_val = None
        return ret_val
    
    def reset(self):
        self._item = None




import asyncio
import inspect
import json
import re
from struct import error
from typing import Callable, Any, Coroutine, Optional

from bclib.utility import DictEx

from bclib.context import ClientSourceContext, RESTfulContext, WebContext, RequestContext, Context, SocketContext, ServerSourceContext, NamedPipeContext
from bclib.listener import Message, MessageType, HttpBaseDataType, ReceiveMessage

from bclib.dispatcher.dispatcher_helper import DispatcherHelper
from bclib.dispatcher.dispatcher import Dispatcher


class RoutingDispatcher(Dispatcher, DispatcherHelper):

    def __init__(self, options: dict,loop:asyncio.AbstractEventLoop=None):
        super().__init__(options=options,loop=loop)
        self.__default_router = self.options.defaultRouter\
            if 'defaultRouter' in self.options and isinstance(self.options.defaultRouter, str)\
            else None
        self.name = self.options["name"] if self.options.has("name") else None
        self.__log_name = f"{self.name}: " if self.name else ''
        if self.options.has('router'):
            router = self.options.router
            if isinstance(router, str):
                self.__context_type_detector: 'Callable[[str],str]' = lambda _: router
            elif isinstance(router, DictEx):
                self.init_router_lookup()
            else:
                raise error(
                    "Invalid value for 'router' property in host options! Use string or dict object only.")
        elif self.__default_router:
            self.__context_type_detector: 'Callable[[str],str]' = lambda _: self.__default_router
        else:
            raise error(
                "Invalid routing config! Please at least set one of 'router' or 'defaultRouter' property in host options.")

    def init_router_lookup(self):
        """create router lookup dictionary"""

        route_dict = dict()
        for key, values in self.options.router.items():
            if key != 'rabbit'.strip():
                if '*' in values:
                    route_dict['*'] = key
                    break
                else:
                    for value in values:
                        if len(value.strip()) != 0 and value not in route_dict:
                            route_dict[value] = key
        if len(route_dict) == 1 and '*' in route_dict and self.__default_router is None:
            router = route_dict['*']
            self.__context_type_detector: 'Callable[[str],str]' = lambda _: router
        else:
            self.__context_type_lookup = route_dict.items()
            self.__context_type_detector = self.__context_type_detect_from_lookup

    def __context_type_detect_from_lookup(self, url: str) -> str:
        """Detect context type from url about lookup"""

        context_type: str = None
        if url:
            try:
                for pattern, lookup_context_type in self.__context_type_lookup:
                    if pattern == "*" or re.search(pattern, url):
                        context_type = lookup_context_type
                        break
            except TypeError:
                pass
            except error as ex:
                print("Error in detect context from routing options!", ex)
        return context_type if context_type else self.__default_router

    async def _on_message_receive_async(self, message: Message) -> Message:
        """Process received message"""

        try:
            context = self.__context_factory(message)
            response = await self.dispatch_async(context)
            ret_val: Message = None
            if context.is_adhoc:
                ret_val = message.create_response_message(
                    message.session_id, 
                    json.dumps(response, ensure_ascii=False).encode("utf-8")
                )
            return ret_val
        except Exception as ex:
            print(f"Error in process received message {ex}")
            raise ex

    def __context_factory(self, message: Message) -> Context:
        """Create context from message object"""

        ret_val: RequestContext = None
        context_type = None
        cms_object: Optional[dict] = None
        url: Optional[str] = None
        request_id: Optional[str] = None
        method: Optional[str] = None
        message_json: Optional[dict] = None
        if message.buffer is not None:
            message_json = json.loads(message.buffer)
            cms_object = message_json[HttpBaseDataType.CMS] if HttpBaseDataType.CMS in message_json else None
            if cms_object:
                if 'request' in cms_object:
                    req = cms_object["request"]
                else:
                    raise KeyError("request key not found in cms object")
                if 'full-url' in req:
                    url = req["full-url"]
                else:
                    raise KeyError("full-url key not found in request")
                request_id = req['request-id'] if 'request-id' in req else 'none'
                method = req['methode'] if 'methode' in req else 'none'
        if message.type == MessageType.AD_HOC:
            if url or self.__default_router is None:
                context_type = self.__context_type_detector(url)
            else:
                context_type = self.__default_router
        else:
            context_type = "socket"
        if self.log_request:
            print(
                f"{self.__log_name}({context_type}::{message.type.name}){f' - {request_id} {method} {url} ' if cms_object else ''}")

        if context_type == "client_source":
            ret_val = ClientSourceContext(cms_object, self)
        elif context_type == "restful":
            ret_val = RESTfulContext(cms_object, self)
        elif context_type == "server_source":
            ret_val = ServerSourceContext(message_json, self)
        elif context_type == "web":
            ret_val = WebContext(cms_object, self,message)
        elif context_type == "socket":
            ret_val = SocketContext(cms_object, self, message, message_json)
        elif context_type == "named_pipe":
            ret_val = NamedPipeContext(message_json, message.buffer.decode("utf-8"), self)
        elif context_type is None:
            raise NameError(f"No context found for '{url}'")
        else:
            raise NameError(
                f"Configured context type '{context_type}' not found for '{url}'")
        return ret_val

    def run_in_background(self, callback: 'Callable|Coroutine', *args: Any) -> asyncio.Future:
        """helper for run function in background thread"""

        if inspect.iscoroutinefunction(callback):
            return self.event_loop.create_task(callback(*args))
        else:
            return self.event_loop.run_in_executor(None, callback, *args)

    async def send_message_async(self, message: MessageType) -> bool:
        """Send message to endpoint"""
        raise NotImplementedError(
            "Send ad-hoc message not support in this type of dispatcher")

    def cache(self, life_time:"int"=0, key:"str"=None):
        """Cache result of function for seconds of time or until signal by key for clear"""

        return self.cache_manager.cache_decorator(key, life_time)



from abc import ABC, abstractmethod
from bclib.context import Context


class Predicate(ABC):
    """Base class for predicate"""

    def __init__(self, expression: str) -> None:
        super().__init__()
        self.exprossion = expression

    @abstractmethod
    async def check_async(self, context: Context) -> bool:
        """Applay cheking for predicate"""



from bclib.utility.http_status_codes import HttpStatusCodes
from .short_circuit_err import ShortCircuitErr


class BadRequestErr(ShortCircuitErr):
    def __init__(self, message: str = None, data: 'dict' = None):
        super().__init__(HttpStatusCodes.BAD_REQUEST, 'http-400', message, data)



import asyncio
import json
from ..logger.schema_base_logger import SchemaBaseLogger
from bclib.utility import DictEx


class RabbitSchemaBaseLogger(SchemaBaseLogger):
    def __init__(self, options: DictEx) -> None:
        super().__init__(options)
        if not options.has("connection") or not options.connection.has("url") or not options.connection.has("queue"):
            raise Exception(
                "connection part of schema logger not set.")

    async def _save_schema_async(self, schema: dict):
        def send_to_rabbit(options):
            import pika
            connection_options = options.connection
            queue = connection_options.queue
            durable = connection_options.durable if connection_options.has("durable") else False
            passive=connection_options.passive if connection_options.has("passive") else False
            exclusive=connection_options.exclusive if connection_options.has("exclusive") else False
            auto_delete=connection_options.auto_delete if connection_options.has("auto_delete") else False
            connection = pika.BlockingConnection(
                pika.URLParameters(connection_options.url))
            channel = connection.channel()
            channel.queue_declare(queue=queue, durable=durable, passive=passive, exclusive=exclusive, auto_delete=auto_delete)
            channel = connection.channel()
            channel.basic_publish(
                exchange='', routing_key=queue, body=json.dumps(schema))
        loop = asyncio.get_running_loop()
        future = loop.run_in_executor(None, send_to_rabbit, self.options)
        await future



class HttpBaseDataName:
    METHODE = "methode"
    RAW_URL = "rawurl"
    DEBUG = "debug"
    HOST = "host"
    COOKIE = "cookie"
    CONTENT_TYPE = "content-type"
    FULL_URL = "full-url"
    PORT = "port"
    URL = "url"
    REQUEST_ID = "request-id"
    HOST_IP = "hostip"
    HOST_PORT = "hostport"
    CLIENT_IP = "clientip"
    DATE = "date"
    DATE2 = "date2"
    DATE3 = "date3"
    TIME = "time"
    TIME2 = "time2"
    BODY = "body"
    HTTP = "http"
    WEB_SERVER = "webserver"
    CONTENT = "content"
    BLOB_CONTENT = "blob-content"
    INDEX = "index"
    HEADER_CODE = "headercode"
    FILE_PATH = "filepath"
    MIME = "mime"
    GZIP = "gzip"



import asyncio
from io import BufferedReader, BufferedWriter
import json
import uuid
from typing import Any
from ..named_pipe.inamed_pipe_connection import INamedPipeConnection

from bclib.utility import LinuxNamedPipeHelper


class LinuxNamedPipeConnection(INamedPipeConnection):
    def __init__(self, pipe_name: str, event_loop: asyncio.AbstractEventLoop) -> None:
        self.__event_loop = event_loop
        self.__reader_pipe_name = F"{pipe_name}-reader"
        self.__writer_pipe_name = F"{pipe_name}-writer"
        self.__reader_pipe_handle: BufferedWriter = None
        self.__writer_pipe_handle: BufferedReader = None
        self.__query_list: 'dict[str,asyncio.Future]' = dict()
        self.__propcess_tesk = self.__event_loop.create_task(
            self.__get_receive_message_Task())

    async def __get_receive_message_Task(self):
        while True:
            try:
                if self.__writer_pipe_handle is None:
                    self.__writer_pipe_handle = open(
                        self.__writer_pipe_name, 'rb')
                message = await LinuxNamedPipeHelper.read_from_named_pipe_async(self.__writer_pipe_handle, self.__event_loop)
                if message and message.session_id in self.__query_list:
                    future = self.__query_list[message.session_id]
                    del self.__query_list[message.session_id]
                    data = json.loads(message.buffer)
                    future.get_loop().call_soon_threadsafe(future.set_result, data)
            except asyncio.CancelledError:
                self.clear_query_list()
                break
            except:
                self.__writer_pipe_handle = None
                self.clear_query_list()
                await asyncio.sleep(1)

    def __try_send_command(self, command: Any) -> str:
        from bclib.listener import Message
        session_id = None
        try:
            if self.__reader_pipe_handle is None:
                self.__reader_pipe_handle = open(self.__reader_pipe_name, "wb")
            session_id = str(uuid.uuid4())
            msg = Message.create_add_hock(
                session_id, json.dumps(command).encode("utf-8"))
            LinuxNamedPipeHelper.write_to_named_pipe(
                msg, self.__reader_pipe_handle)
        except:
            self.__reader_pipe_handle = None
            raise
        return session_id

    def try_send_command(self, command: Any) -> bool:
        return True if self.__try_send_command(command) else False

    async def try_send_query_async(self, query: 'Any') -> 'Any':
        ret_val = None
        session_id = self.__try_send_command(query)
        if session_id:
            future = self.__event_loop.create_future()
            self.__query_list[session_id] = future
            ret_val = await future
        return ret_val

    def clear_query_list(self):
        for session_id in self.__query_list:
            try:
                future = self.__query_list[session_id]
                future.get_loop().call_soon_threadsafe(
                    future.set_exception, Exception("Named Pipe is broken"))
            except:
                pass
        self.__query_list.clear()



import sqlite3
from ..db_manager.db import Db


class SQLiteDb(Db):
    """SQLite implementation of Db wrapper"""

    def __init__(self, connection_string: str) -> None:
        super().__init__()
        self.connection = sqlite3.connect(connection_string)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()
        return super().__exit__(exc_type, exc_val, exc_tb)



from ..cache.manager import CacheManager
from abc import ABC
from bclib.utility import DictEx
from ..cache.in_memory_cache_manager import InMemoryCacheManager
from ..cache.no_cache import NoCacheManager

class CacheFactory(ABC):
    @staticmethod
    def create(options:"DictEx"=None) -> "CacheManager":
        cache_type = str(options.type) if options is not None and options.has("type") else None
        if cache_type is not None:
            if cache_type == "memory":
                return InMemoryCacheManager(options)
            else:
                raise ValueError(f"Unknown type for cache ('${cache_type}')")
        return NoCacheManager(options)


from bclib.utility.http_status_codes import HttpStatusCodes
from ..exception.short_circuit_err import ShortCircuitErr


class UnauthorizedErr(ShortCircuitErr):
    def __init__(self, message: str = None, data: 'dict' = None):
        super().__init__(HttpStatusCodes.UNAUTHORIZED, 'http-401', message, data)



import asyncio
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from bclib.listener.message import Message


class WindowsNamedPipeHelper:

    @staticmethod
    def __check_read_error(error_code):
        if error_code != 0:
            raise Exception(
                f"error in read from pip. error code = {error_code}")

    @staticmethod
    def __check_write_error(error_code):
        if error_code != 0:
            raise Exception(
                f"error in write in pip. error code = {error_code}")

    @staticmethod
    def __read_from_named_pipe(named_pipe_handle: Any, future: asyncio.Future) -> None:
        import win32file
        from bclib.listener import Message, MessageType

        try:
            error, data = win32file.ReadFile(named_pipe_handle, 1)
            WindowsNamedPipeHelper.__check_read_error(error)
            message_type = MessageType(int.from_bytes(
                data, byteorder='big', signed=True))
            error, data = win32file.ReadFile(named_pipe_handle, 4)
            WindowsNamedPipeHelper.__check_read_error(error)
            data_len = int.from_bytes(
                data, byteorder='big', signed=True)
            error, data = win32file.ReadFile(named_pipe_handle, data_len)
            WindowsNamedPipeHelper.__check_read_error(error)
            session_id = data.decode("utf-8")
            parameter = None
            if message_type in (MessageType.AD_HOC, MessageType.MESSAGE, MessageType.CONNECT):
                error, data = win32file.ReadFile(named_pipe_handle, 4)
                WindowsNamedPipeHelper.__check_read_error(error)
                data_len = int.from_bytes(
                    data, byteorder='big', signed=True)
                error, parameter = win32file.ReadFile(
                    named_pipe_handle, data_len)
                WindowsNamedPipeHelper.__check_read_error(error)
        except Exception as ex:
            future.get_loop().call_soon_threadsafe(future.set_exception, ex)
        else:
            message = Message(session_id, message_type, parameter)
            future.get_loop().call_soon_threadsafe(future.set_result, message)

    @staticmethod
    async def read_from_named_pipe_async(named_pipe_handle: Any, loop: asyncio.AbstractEventLoop = None) -> 'Message':
        if not loop:
            loop = asyncio.get_event_loop()
        future = loop.create_future()
        loop.run_in_executor(
            None, WindowsNamedPipeHelper.__read_from_named_pipe, named_pipe_handle, future)
        return await future

    @staticmethod
    def write_to_named_pipe(message: 'Message', named_pipe_handle: Any):
        import win32file
        import win32pipe
        import pywintypes
        from bclib.listener import MessageType

        try:
            error, _ = win32file.WriteFile(
                named_pipe_handle, message.type.value.to_bytes(1, 'big'))
            WindowsNamedPipeHelper.__check_write_error(error)
            data = message.session_id.encode()
            data_length_bytes = len(data).to_bytes(4, 'big')
            error, _ = win32file.WriteFile(
                named_pipe_handle, data_length_bytes)
            WindowsNamedPipeHelper.__check_write_error(error)
            error, _ = win32file.WriteFile(named_pipe_handle, data)
            WindowsNamedPipeHelper.__check_write_error(error)
            if message.type in (MessageType.AD_HOC, MessageType.MESSAGE):
                data_length_bytes = len(message.buffer).to_bytes(4, 'big')
                error, _ = win32file.WriteFile(
                    named_pipe_handle, data_length_bytes)
                WindowsNamedPipeHelper.__check_write_error(error)
                error, _ = win32file.WriteFile(
                    named_pipe_handle, message.buffer)
                WindowsNamedPipeHelper.__check_write_error(error)
            win32file.FlushFileBuffers(named_pipe_handle)
        except pywintypes.error as e:  # pylint: disable=maybe-no-member
            # Disconnect the named pipe
            win32pipe.DisconnectNamedPipe(named_pipe_handle)
            # CLose the named pipe
            win32file.CloseHandle(named_pipe_handle)
            if e.args[0] == 2:   # ERROR_FILE_NOT_FOUND
                print(f"No Named Pipe.  {repr(e)}")
            elif e.args[0] == 232:
                print(f"The Pipe is being closed. {repr(e)}")
            elif e.args[0] == 109:   # ERROR_BROKEN_PIPE
                print(f"Named Pipe is broken. {repr(e)}")
            else:
                print(f"Named Pipe error code {e.args[0]}. {repr(e)}")
            raise

    @staticmethod
    async def wait_for_client_connect_async(pipe_handler: any, loop: asyncio.AbstractEventLoop = None) -> None:
        import win32pipe
        future = loop.create_future()

        def process(pipe_handler: any, future: asyncio.Future):
            try:
                win32pipe.ConnectNamedPipe(pipe_handler, None)
                future.get_loop().call_soon_threadsafe(future.set_result, True)
            except Exception as ex:
                future.get_loop().call_soon_threadsafe(future.set_exception, ex)
        if not loop:
            loop = asyncio.get_event_loop()
        loop.run_in_executor(None, process, pipe_handler, future)
        await future



from typing import Any, Coroutine, Optional
from aiohttp import web
from bclib.listener.message import Message
from bclib.listener.message_type import MessageType

class WebMessage(Message):
    def __init__(self,  request: 'web.Request',sessionId: str,  messageType: MessageType, buffer: bytes = None) -> None:
        super().__init__(sessionId,messageType,buffer)
        self.__request = request
        self.Response = None


    def create_response_message(self, session_id: str, buffer: bytes) -> "Message":
        ret_val = WebMessage (self.__request, session_id,MessageType.AD_HOC,buffer)
        ret_val.Response = self.Response
        return ret_val

    async def start_stream_response_async(self,status: int = 200,
        reason: Optional[str] = 'OK',
        headers: Optional[dict] = None,):
        if self.Response is not None:
            raise Exception('StreamResponse started')
        self.Response = web.StreamResponse(status=status, 
                                    reason=reason, 
                                    headers=headers)
        await self.Response.prepare(self.__request)

    async def write_async(self,data:'bytes') -> Coroutine[Any, Any, None]:
        if self.Response is None:
            raise Exception('StreamResponse not started')
        await self.Response.write(data)

    async def  drain_async(self) -> Coroutine[Any, Any, None]:
        if self.Response is None:
            raise Exception('StreamResponse not started')
        await self.Response.drain()


from typing import Any
from ..context.merge_type import MergeType
from ..context.context import Context
from ..context.server_source_context import ServerSourceContext


class ServerSourceMemberContext(Context):
    """Context for Server dbSource member request"""

    def __init__(self, sourceContext: ServerSourceContext, data: Any, member: dict) -> None:
        super().__init__(sourceContext.dispatcher)
        self.__source_context = sourceContext
        self.member = member
        self.data = data
        self.command = self.__source_context.command
        self.table_name: str = f"{sourceContext.command.name}.{member.name}"
        self.key_field_name: str = None
        self.status_field_name: str = None
        self.merge_type: MergeType = MergeType.REPLACE
        self.column_names: 'list[str]' = None



from abc import ABC
import re


class Validator(ABC):

    @staticmethod
    def check_validators(validators:"dict", value:"any") -> "tuple[bool, list[str]]":

        check_validation = {
            "required": Validator.required_validator(value, validators["required"]) if 'required' in validators else None,
            "minLength": Validator.min_length_validator(value, validators["minLength"]) if validators.get('minLength') is not None else None,
            "maxLength": Validator.max_length_validator(value, validators["maxLength"]) if validators.get('maxLength') is not None else None, 
            "min": Validator.min_validator(value, validators["min"]) if validators.get('min') is not None else None, 
            "max": Validator.max_validator(value, validators["max"]) if validators.get('max') is not None else None, 
            "dataType": Validator.data_type_validator(value, validators["dataType"]) if 'dataType' in validators else None,
            "regex": Validator.regex_validator(value, validators["regex"]) if 'regex' in validators else None
        }
        validations_message:"list[str]" = []
        
        for validator in validators:
            if validator in check_validation:
                checked = check_validation[validator]
                if checked is not None:
                    checked_status, checked_message = checked
                    if not checked_status:
                        validations_message.append(checked_message)

        validation_status = True if len(validations_message) == 0 else False
        return validation_status, validations_message
    
    @staticmethod
    def required_validator(value:"any", is_required:"bool") -> "tuple[bool, list[str]|None]":
        status = False if is_required and not value else True
        message = None if status else "Required validation error"
        return status, message

    @staticmethod
    def min_length_validator(value:"any", min_length:"any") -> "tuple[bool, list[str]|None]":
        try:
            min_length = min_length if isinstance(min_length, int) else int(min_length)
            status = len(value) >= min_length
        except Exception:
            status = False

        message = None if status else "Min length error"
        return status, message

    @staticmethod
    def max_length_validator(value:"any", max_length:"any") -> "tuple[bool, list[str]|None]":
        try:
            max_length = max_length if isinstance(max_length, int) else int(max_length)
            status = len(value) <= max_length
        except Exception:
            status = False

        message = None if status else "Max length error"
        return status, message
    
    @staticmethod
    def min_validator(value:"any", min_value:"int") -> "tuple[bool, list[str]|None]":
        try:
            status = int(value) >= min_value
        except Exception as ex:
            status = False
        
        message = None if status else "Min value error"
        return status, message
    
    @staticmethod
    def max_validator(value:"any", max_value:"int") -> "tuple[bool, list[str]|None]":
        try:
            status = int(value) <= max_value
        except Exception as ex:
            status = False
        
        message = None if status else "Max value error"
        return status, message 
    
    @staticmethod
    def data_type_validator(value:"any", type:"str") -> "tuple[bool, list[str]|None]":        
        message = None
        status = True
        if type in ["int", "float", "text", "ntext", "str"]:
            try:
                if type == "int":
                    int(value)
                elif type == "float":
                    float(value)
                else:
                    str(value)
            except Exception:
                status = False
            if not status:
                message = f"Type of value is not {type}"
        else:
            status = False
            message = f"Function not found for {type} type checker"

        return status, message

    @staticmethod
    def regex_validator(value:"any", regex_expression:"str") -> "tuple[bool, list[str]|None]":
        try:
            status = re.match(regex_expression, value.strip()) != None
        except Exception as ex:
            status = False
        
        message = None if status else "regex error"
        return status, message


from bclib.db_manager.named_pipe.named_pipe_connection import NamedPipeConnection
from bclib.db_manager.named_pipe.inamed_pipe_connection import INamedPipeConnection
from bclib.db_manager.named_pipe.windows_named_pipe_connection import WindowsNamedPipeConnection



from abc import ABC, abstractmethod
from typing import Any


class INamedPipeConnection(ABC):

    @abstractmethod
    def try_send_command(self, command: 'Any') -> bool:
        pass

    @abstractmethod
    async def try_send_query_async(self, query: 'Any') -> 'Any':
        pass



from bclib.parser.answer.answer import Answer
from bclib.parser.answer.user_action_types import UserActionTypes
from bclib.parser.answer.user_action import UserAction
from bclib.parser.answer.file_user_action import FileUserAction



"""Dispatcher base class module"""
from abc import ABC, abstractmethod
import asyncio
from typing import Callable, Any, TYPE_CHECKING, Coroutine

from bclib.db_manager import DbManager
from bclib.cache import CacheManager
from bclib.listener import Message
from bclib.utility import DictEx

if TYPE_CHECKING:
    from context import Context


class IDispatcher(ABC):
    """Dispatcher base class with core functionality for manage cache and background process"""

    @property
    @abstractmethod
    def log_error(self) -> bool:
        pass

    @property
    @abstractmethod
    def log_request(self) -> bool:
        pass

    @property
    @abstractmethod
    def event_loop(self) -> asyncio.AbstractEventLoop:
        pass

    @property
    @abstractmethod
    def db_manager(self) -> DbManager:
        pass

    @property
    @abstractmethod
    def options(self) -> DictEx:
        pass

    @property
    @abstractmethod
    def cache_manager(self) -> CacheManager:
        pass

    @abstractmethod
    async def dispatch_async(self, context: 'Context') -> Any:
        """Dispatch context and get result from related action method"""

    @abstractmethod
    def dispatch_in_background(self, context: 'Context') -> asyncio.Future:
        """Dispatch context in background"""

    @abstractmethod
    async def send_message_async(self, message: Message) -> None:
        """Send message to endpoint"""

    def run_in_background(self, callback: Callable, *args: Any) -> asyncio.Future:
        """helper for run function in background thread"""

    def log_async(self, **kwargs: 'dict[str,Any]'):
        """log params"""

    def log_in_background(self, **kwargs: 'dict[str,Any]') -> Coroutine:
        """log params in background precess"""



from abc import ABC, abstractmethod
from typing import Any

from ..context.merge_type import MergeType

from bclib.utility import DictEx


class SourceMemberContext(ABC):

    @property
    @abstractmethod
    def member(self) -> DictEx:
        pass

    @property
    @abstractmethod
    def data(self) -> Any:
        pass

    @property
    @abstractmethod
    def command(self) -> DictEx:
        pass

    @property
    @abstractmethod
    def table_name(self) -> str:
        pass

    @property
    @abstractmethod
    def key_field_name(self) -> str:
        pass

    @property
    @abstractmethod
    def status_field_name(self) -> str:
        pass

    @property
    @abstractmethod
    def merge_type(self) -> MergeType:
        pass

    @property
    @abstractmethod
    def column_names(self) -> 'list[str]':
        pass



from bclib.exception.short_circuit_err import ShortCircuitErr
from bclib.exception.unauthorized_err import UnauthorizedErr
from bclib.exception.internal_server_err import InternalServerErr
from bclib.exception.not_found_err import NotFoundErr
from bclib.exception.handler_not_found_err import HandlerNotFoundErr
from bclib.exception.bad_request_err import BadRequestErr
from bclib.exception.forbidden_err import ForbiddenErr


import json
from typing import TYPE_CHECKING
from bclib.utility import DictEx
from ..context.json_base_request_context import JsonBaseRequestContext
from urllib.parse import parse_qsl

if TYPE_CHECKING:
    from .. import dispatcher


class RESTfulContext(JsonBaseRequestContext):
    def __init__(self, cms_object: dict, dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(cms_object, dispatcher)
        temp_data = None
        if self.cms.form:
            temp_data = self.cms.form
        else:
            body = self.cms.request.body
            if body:
                content_type = self.cms.request.get("content-type")
                if content_type.find("x-www-form-urlencoded") > -1:
                    temp_data = dict()
                    for key, value in parse_qsl(body):
                        temp_data[key.strip()] = value
                else:
                    try:
                        temp_data = json.loads(self.cms.request.body)
                    except Exception as ex:
                        print('error in extract request body', ex)
        self.body = DictEx(temp_data) if temp_data else None
